/*
This work is licensed under a Creative Commons Attribution-NonCommercial-ShareAlike 4.0 International License.
http://creativecommons.org/licenses/by-nc-sa/4.0/legalcode
Initial credits go to Nick Gammon for serial state machine

Philosophy:
1- decode serial communication using state machine
2- Manage update of global value including controls and math if needed (process family functions)
3- Update display using handlePreviousState
*/

/*
Serial syntax: [command letter][decimal number]
// Yx      : Start module with intensity set to x (1-7)
// Sx      : Speed update (decimal notation)
// Gx      : Gear update (decimal notation)
// Rx      : RPM update (decimal notation)
// Ox      : oil temp
// Wx      : water temp
// Fx      : fuel
// Lx      : lap
// Px      : position
// Z0A     : Stop and clear module

The command will be processed when a new one starts. For example, RPM will be updated on the module once a new command such as S will start
Serial data to start with max luminosity set (5), speed=95, RPM=4250, gear=2 (last caracter is to finish processing of RPM, can be anything except a digit)
Y5S95G2R4250R

If gear is negative, the reverse is assumed and 'r' is displayed
If gear is 0, neutral is assumed and 'n' is displayed

Usage
button1 : cycle values for left dashboard part
button2 : cycle values for right dashboard part
button3 : toggle speed multiplier by 3.6 (needed by some games such as Kart Racing Pro. Can be done externally also in dash setting)
button8 : show current settings and save to EEPROM. Third digit shown indicates if speed multiplier is on (1) or off (0)
*/

//#define DEBUG
#include <TM1638.h>
#include <EEPROMex.h>
#undef _EEPROMEX_DEBUG //to prevent false alarms on EEPROM write execeeded
#define LEFT 1
#define RIGHT 2
unsigned int carspeed;         // holds the speed data (0-65535 size)
unsigned int cargear;          // holds gear value data (0-65535 size)
unsigned int carrpm;           // holds the rpm data (0-65535 size)
unsigned int caroil;           // holds the oil temp data (0-65535 size)
unsigned int carwater;         // holds the water temp data (0-65535 size)
unsigned int carfuel;          // holds the fuel data (0-65535 size)
unsigned int carlap;           // holds the lap data (0-65535 size)
unsigned int carpos;           // holds the position data (0-65535 size)

unsigned int rpmleds;          // holds the 8 leds values, both MSB=green LSB=red
unsigned int rpmmax = 5000;    // autocalibrating RPM max val
bool speedmultiplier = 0;      //speed multiplier by 3.6 (usefull for some games such as kart racing pro)
// buttons TM1638 (left to right)
#define btn1 0b00000001 // 1
#define btn2 0b00000010 // 2
#define btn3 0b00000100 // 4
#define btn4 0b00001000 // 8
#define btn5 0b00010000 // 16
#define btn6 0b00100000 // 32
#define btn7 0b01000000 // 64
#define btn8 0b10000000 // 128

const byte indicator[] = { 'S','G','R','O','W','F','L' }; //possible indicators show on module
byte leftindicator = 0; //speed on left by default
byte rightindicator = 1; //gear on right by default
byte buttons;
bool reverse;
String text;
int offset;
// the possible states of the state-machine
typedef enum { NONE, GOT_R, GOT_S, GOT_G, GOT_O, GOT_W, GOT_F, GOT_L, GOT_P, GOT_STOP, GOT_START } states;

// current state-machine state
states state = NONE;
// current partial number
signed int currentValue;
//Data Clock Strobe
TM1638 module1(5, 4, 3);

void processRPM(const unsigned int value)
{
	// do something with RPM
#if defined DEBUG
	Serial.print(F("RPM = ")); Serial.println(value);
#endif
	carrpm = value;
	if (carrpm > rpmmax && carrpm < 50000) {
		rpmmax = carrpm;
	}
}

void processSpeed(const unsigned int value)
{
#if defined DEBUG
	Serial.print(F("Speed = ")); Serial.println(value);
#endif
	if (speedmultiplier) carspeed = value * 3.6;
	else carspeed = value;
}

void processGear(const unsigned int value)
{
	// do something with gear
#if defined DEBUG
	Serial.print(F("Gear = ")); Serial.println(value);
#endif
	cargear = value;
}

void processOil(const unsigned int value)
{
#if defined DEBUG
	Serial.print(F("Oil = ")); Serial.println(value);
#endif
	caroil = value;
}

void processWater(const unsigned int value)
{
#if defined DEBUG
	Serial.print(F("Water = ")); Serial.println(value);
#endif
	carwater = value;
}

void processFuel(const unsigned int value)
{
#if defined DEBUG
	Serial.print(F("Fuel = ")); Serial.println(value);
#endif
	carfuel = value;
}

void processLap(const unsigned int value)
{
#if defined DEBUG
	Serial.print(F("Lap = ")); Serial.println(value);
#endif
	carlap = value;
}

void processPos(const unsigned int value)
{
#if defined DEBUG
	Serial.print(F("Pos = ")); Serial.println(value);
#endif
	carpos = value;
}

void processStop(const unsigned int value)
{
#if defined DEBUG
	Serial.print(F("Stop = ")); Serial.println(value);
#endif
	module1.clearDisplay();
	module1.setLEDs(0);
	module1.setupDisplay(false, 7);
}

void processStart(const unsigned int value)
{
#if defined DEBUG
	Serial.print(F("Start = ")); Serial.println(value);
#endif
	module1.setupDisplay(true, value);
	module1.clearDisplay();
	module1.setLEDs(0);
	rpmmax = 5000;
}

void printModule(unsigned int value, int pad, int side) {
	text = String(value);
	while (text.length() < pad) {
		text = "0" + text;
	}
	if (side == LEFT) {// we print on the left side of the module
#if defined DEBUG
		Serial.print(F("Change left to: ")); Serial.print(text); Serial.println("");
#endif
		offset = 0;
	}
	else {// we print on the right side of the module
#if defined DEBUG
		Serial.print(F("Change right to: ")); Serial.print(text); Serial.println("");
#endif
		offset = 8 - text.length();
	}
	module1.setDisplayToString(text, 0, offset);
}

void handlePreviousState()
{
	switch (state)
	{
	case GOT_R:
		processRPM(currentValue);
		if (indicator[leftindicator] == 'R') {
			printModule(carrpm, 4, LEFT);
		}
		if (indicator[rightindicator] == 'R') {
			printModule(carrpm, 4, RIGHT);
		}
		//Update LEDS
		rpmleds = map(carrpm, 0, rpmmax * 1.1, 0, 10);        // distributes the rpm level to the 8 leds + 1 for shift change
#if defined DEBUG
		Serial.println(rpmleds);
#endif
		if (rpmleds == 0) {
			module1.setLEDs(0x0000);
		}
		if (rpmleds == 1) {
			module1.setLEDs(0x0001);
		}
		if (rpmleds == 2) {
			module1.setLEDs(0x0003);
		}
		if (rpmleds == 3) {
			module1.setLEDs(0x0007);
		}
		if (rpmleds == 4) {
			module1.setLEDs(0x000F);
		}
		if (rpmleds == 5) {
			module1.setLEDs(0X001F);
		}
		if (rpmleds == 6) {
			module1.setLEDs(0X201F);
		}
		if (rpmleds == 7) {
			module1.setLEDs(0X601F);
		}
		if (rpmleds == 8) {
			module1.setLEDs(0XE01F);
		}
		if (rpmleds == 9) {
			module1.setLEDs(0XFF00);
		}
		break;
	case GOT_W:
		processWater(currentValue);
		if (indicator[leftindicator] == 'W') {
			printModule(carwater, 3, LEFT);
		}
		if (indicator[rightindicator] == 'W') {
			printModule(carwater, 3, RIGHT);
		}
		break;
	case GOT_F:
		processFuel(currentValue);
		if (indicator[leftindicator] == 'F') {
			printModule(carfuel, 3, LEFT);
		}
		if (indicator[rightindicator] == 'F') {
			printModule(carfuel, 3, RIGHT);
		}
		break;
	case GOT_O:
		processOil(currentValue);
		if (indicator[leftindicator] == 'O') {
			printModule(caroil, 3, LEFT);
		}
		if (indicator[rightindicator] == 'O') {
			printModule(caroil, 3, RIGHT);
		}
		break;
	case GOT_S:
		processSpeed(currentValue);
		if (indicator[leftindicator] == 'S') {
			printModule(carspeed, 3, LEFT);
		}
		if (indicator[rightindicator] == 'S') {
			printModule(carspeed, 3, RIGHT);
		}
		break;
	case GOT_G:
		processGear(currentValue);
		if (indicator[leftindicator] == 'G') {
			if (reverse == false) {
				if (cargear == 0) {
#ifdef DEBUG
					Serial.println(F("Neutral left"));
#endif
					module1.setDisplayToString("n", 0, 0);
				}
				else { printModule(cargear, 1, LEFT); }
			}
			else {
#ifdef DEBUG
				Serial.println(F("Reverse left"));
#endif
				module1.setDisplayToString("R", 0, 0);
			}
		}
		if (indicator[rightindicator] == 'G') {
			if (reverse == false) {
				if (cargear == 0) {
#ifdef DEBUG
					Serial.println(F("Neutral right"));
#endif
					module1.setDisplayToString("n", 0, 7);
				}
				else { printModule(cargear, 1, RIGHT); }
			}
			else {
#ifdef DEBUG
				Serial.println(F("Reverse right"));
#endif
				module1.setDisplayToString("R", 0, 7);
			}
		}
		reverse = false;
		break;
	case GOT_L:
		processLap(currentValue);
		if (indicator[leftindicator] == 'L') {
			printModule(carlap, 2, LEFT);
		}
		if (indicator[rightindicator] == 'L') {
			printModule(carlap, 2, RIGHT);
		}
		break;
	case GOT_P:
		processPos(currentValue);
		if (indicator[leftindicator] == 'P') {
			printModule(carpos, 2, LEFT);
		}
		if (indicator[rightindicator] == 'P') {
			printModule(carpos, 2, RIGHT);
		}
		break;
	case GOT_START:
		processStart(currentValue);
		break;
	case GOT_STOP:
		processStop(currentValue);
		break;
	}  // end of switch
	currentValue = 0;
}  // end of handlePreviousState

void processIncomingByte(const byte c)
{
	if (isdigit(c))
	{
		currentValue *= 10;
		currentValue += c - '0';
	}  // end of digit

	else {
		// The end of the number signals a state change
		if (c != '-') { handlePreviousState(); } //we need to manage minus sign for negative gear, meaning reverse speed
	  // set the new state, if we recognize it
		switch (c)
		{
		case '-':
#ifdef DEBUG
			Serial.println(F("Minus car for reverse gear"));
#endif
			reverse = true;
			state = GOT_G;
			break;
		case 'R':
			state = GOT_R;
			break;
		case 'S':
			state = GOT_S;
			break;
		case 'G':
			state = GOT_G;
			break;
		case 'O':
			state = GOT_O;
			break;
		case 'W':
			state = GOT_W;
			break;
		case 'F':
			state = GOT_F;
			break;
		case 'L':
			state = GOT_L;
			break;
		case 'P':
			state = GOT_P;
			break;
		case 'Z':
			state = GOT_STOP;
			break;
		case 'Y':
			state = GOT_START;
			break;
		default:
			state = NONE;
			break;
		}  // end of switch on incoming byte
	} // end of not digit

} // end of processIncomingByte

void dispIndicators()
{
	char l;
	char r;
	l = (char)(indicator[leftindicator]);
	r = (char)(indicator[rightindicator]);
	text = String(l) + " " + String(speedmultiplier) + "    " + String(r);
#if defined DEBUG
	Serial.print(F("Indicators:")); Serial.println(text);
#endif
	module1.clearDisplay();
	module1.setDisplayToString(text);
	delay(200);
}

/*void updateLEDs(unsigned int rpm) {
  // light the 3rd red LED
  module1.setLED(TM1638_COLOR_RED, 3);
  // light the 5th green LED
  module1.setLED(TM1638_COLOR_GREEN, 5);
  // light the 7th red and green LEDs
  module1.setLED(TM1638_COLOR_RED | TM1638_COLOR_GREEN, 7);
}
*/

void setup()
{
	Serial.begin(115200);
	state = NONE;
	module1.setupDisplay(true, 7);
	module1.clearDisplay();
	module1.setDisplayToString(F("CarDashX"));
	module1.setLEDs(0x00ff);
	delay(1000);
	leftindicator = EEPROM.readByte(5);
	rightindicator = EEPROM.readByte(7);
	speedmultiplier = EEPROM.readByte(9);
	dispIndicators();
	module1.setLEDs(0x00);
	reverse = false;
}  // end of setup

void loop()
{
	//  buttons = btn4; //For testing purpose
	while (1 == 1) {
		while (Serial.available()) processIncomingByte(Serial.read());
		buttons = module1.getButtons();
		if (buttons != 0 && buttons != 255) {
			//      Serial.print ("Button = "); Serial.println (buttons);
			switch (buttons)
			{
			case btn1:
				if (leftindicator + 1 < sizeof(indicator)) leftindicator++;
				else leftindicator = 0;
#if defined DEBUG
				Serial.print(F("Change left to ")); Serial.println((char)indicator[leftindicator]);
#endif
				dispIndicators();
				delay(1000);
				module1.clearDisplay();
				//buttons = btn2; //For testing purpose
				break;
			case btn2:
				if (rightindicator + 1 < sizeof(indicator)) rightindicator++;
				else rightindicator = 0;
#if defined DEBUG
				Serial.print(F("Change right to ")); Serial.println((char)indicator[rightindicator]);
#endif
				dispIndicators();
				delay(1000);
				module1.clearDisplay();
				//buttons = btn3; //For testing purpose
				break;
			case btn3:
				speedmultiplier = !speedmultiplier;
				dispIndicators();
				delay(1000);
				module1.clearDisplay();
#if defined DEBUG
				Serial.print(F("Speed multiplier set ")); Serial.println(speedmultiplier);
#endif
				//buttons = btn4;
				break;
			case btn8:
				dispIndicators();
				EEPROM.updateByte(5, leftindicator);
				EEPROM.updateByte(7, rightindicator);
				EEPROM.updateByte(9, speedmultiplier);
#if defined DEBUG
				Serial.print(F("EEPROM updated ")); Serial.print(EEPROM.readByte(5)); Serial.print(EEPROM.readByte(7)); Serial.print(EEPROM.readByte(9));
#endif
				//buttons = 0;
				delay(2000);
				module1.clearDisplay();
				break;
			}
			delay(200);//debounce
		}
	}
}
